import torch
from VerboseLayer import VerboseLayer

class CNN(torch.nn.Module):
    def __init__(self, channels, linear_layers, kernel_size=3, stride=1, dropout=0.5, verbose=False, activate_last_linear=False):
        super().__init__()
        l = []

        for idx in range(len(channels) - 1):
            l.append(
                torch.nn.Dropout(dropout)
            )
            l.append(
                torch.nn.Conv2d(channels[idx], channels[idx + 1], kernel_size, stride)                    
            )
            if verbose:
                l.append(VerboseLayer(idx+1))
            l.append(
                torch.nn.BatchNorm2d(channels[idx + 1])
            )
            l.append(torch.nn.ReLU())   # activate
        for idx in range(len(linear_layers) - 1):
            if idx == 0:
                l.append(torch.nn.Flatten())
            l.extend([
                torch.nn.Linear(linear_layers[idx], linear_layers[idx + 1])
            ])
            if activate_last_linear or idx + 1 != len(linear_layers): # if this is not the last layer ( +1 = zero indexed) (-1 = layer b4 last)
                l.append(torch.nn.ReLU())   # activate

        self.layers = torch.nn.Sequential(
            *l
        )

    def forward(self, x):
        x = self.layers(x)
        return x


import torch
from VerboseLayer import VerboseLayer

class Unflatten(torch.nn.Module):
    def __init__(self, size):
        super().__init__()
        self.size = size

    def forward(self, x):
        return x.view(x.shape[0], *self.size)

class TransposeCNN(torch.nn.Module):
    def __init__(self, 
            channels, 
            linear_layers, 
            kernel_size=3, 
            stride=1, 
            dropout=0.5, 
            output_padding=None,
            unflattened_size=None, # required to unflatten to a correct size
            verbose=False
        ):
        super().__init__()
        l = []

        for idx in range(len(linear_layers) - 1):
            l.extend([
                torch.nn.Linear(linear_layers[idx], linear_layers[idx + 1])
            ])
            l.append(torch.nn.ReLU())   # activate
            # print(idx, len(linear_layers) - 1)
            if idx + 1 == len(linear_layers) - 1: # at the end of the linear layers
                print(unflattened_size)
                l.append(Unflatten(unflattened_size))# unflatten the example

        for idx in range(len(channels) - 1):
            l.append(
                torch.nn.Dropout(dropout)
            )
            if output_padding:
                out_pad = output_padding[idx] # the plus one is to offset the padding. padding remainders are generated from the input layers of the conv, but need to be applied to the output of the transpose conv
            else:
                out_pad = 0
            l.append(
                torch.nn.ConvTranspose2d(channels[idx], channels[idx + 1], kernel_size, stride, output_padding=out_pad)
            )
            if verbose:
                l.append(VerboseLayer(idx))
            l.append(
                torch.nn.BatchNorm2d(channels[idx + 1])
            )
            if idx + 1 != len(channels): # if this is not the last layer ( +1 = zero indexed) (-1 = layer b4 last)
                l.append(torch.nn.ReLU())   # activate
            else:
                l.append(torch.nn.Sigmoid()) # apply sigmoid to put output pixel intensities in real range
 
        self.layers = torch.nn.Sequential(
            *l
        )

    def forward(self, x):
        x = self.layers(x)
        return x


from PIL import Image
from PIL import ImageFile
import torch
import torchvision.transforms as transforms
import glob
import os
import random
import json
import requests
from pathlib import Path
from tqdm import tqdm

ImageFile.LOAD_TRUNCATED_IMAGES = True

def tmp_func(x):
    return x.repeat(3, 1, 1)

class Furniture(torch.utils.data.Dataset):
    '''
    The ImageDataset object inherits its methods from the
    torch.utils.data.Dataset module.
    It loads all images from an image folder, creating labels
    for each image according to the subfolder containing the image
    and transform the images so they can be used in a model
    Parameters
    ----------
    root_dir: str
        The directory with subfolders containing the images
    transform: torchvision.transforms 
        The transformation or list of transformations to be 
        done to the image. If no transform is passed,
        the class will do a generic transformation to
        resize, convert it to a tensor, and normalize the numbers
    
    Attributes
    ----------
    files: list
        List with the directory of all images
    labels: set
        Contains the label of each sample
    dict_encoder: dict
        Dictionary to translate the label to a 
        numeric value
    '''
    def __init__(self, root_dir, transform=True, download=True):
        
        self.root_dir = root_dir
        if download:
            self.download(self.root_dir)
        else:
            if not os.path.exists(root_dir):
                raise RuntimeError('Dataset not found.' +
                                   'You can use download=True to download it')

        self.files = glob.glob(os.path.join(f'{self.root_dir}', '*', '*', '*' + '.*'))
        self.labels = list(set([fp.split(os.sep)[1] for fp in self.files]))
        self.num_classes = len(self.labels)
        self.dict_encoder = {y: x for (x, y) in enumerate(self.labels)}
        self.transform = transform
        if transform is True:
            self.transform = transforms.Compose([
                transforms.Resize(64),
                transforms.CenterCrop(64),
                transforms.ToTensor(),
                transforms.Normalize((0.5,), (0.5,)) # GANHACK #1
            ])
            
            self.transform_Gray = transforms.Compose([
            transforms.Resize(64),
            transforms.CenterCrop(64),
            transforms.ToTensor(),
            transforms.Lambda(tmp_func),
            transforms.Normalize((0.5,), (0.5,))
            ])

    def __getitem__(self, index):

        img_name = self.files[index]
        label = img_name.split(os.sep)[1]
        label = self.dict_encoder[label]
        label = torch.as_tensor(label)
        image = Image.open(img_name)
        if self.transform:
            image = self.transform(image)
        return image, label

    def __len__(self):
        return len(self.files)
    
    def download(self, root):

        with open('DATA/data.json') as f:
            data = json.load(f) # read data containing image paths

        paths = ('/'.join(path.split('/')[1:]) for category in data.values() for item in category.values() for path in item['images']) # generate paths

        for path in tqdm(paths):
            # OS FRIENDLY WAYS TO GET THE IMG PATH AND DIR
            fp = os.path.join(self.root_dir, *os.path.split(path))
            if os.path.exists(fp):
                continue
            dir = os.path.join(*os.path.split(fp)[:1])
            Path(dir).mkdir(parents=True, exist_ok=True) # create dir if doesnt exist
            response = requests.get(f'https://ikea-dataset.s3.us-east-2.amazonaws.com/data/{path}')
            with open(fp, 'wb') as f:
                f.write(response.content)

def split_train_test(dataset, train_percentage):
    train_split = int(len(dataset) * train_percentage)
    train_dataset, validation_dataset = torch.utils.data.random_split(
        dataset, [train_split, len(dataset) - train_split]
    )
    return train_dataset, validation_dataset

if __name__ == '__main__':
    dataset = Furniture(root_dir='images', download=True, transform=False)

    # GIVE RANDOM EXAMPLE
    while True:
        idx = random.randint(0, len(dataset))
        img, label = dataset[idx]
        print(label)
        img.show()
        break # your computer will die if you remove this


#%%
import sys
sys.path.append('../../Furniture-StyleGAN')
from ImageDataset import ImageDataset
import torch
import matplotlib.pyplot as plt
import numpy as np
from torch import nn
import torchvision
from time import sleep

batch_size = 32
nc = 3 # Number of channels in the training images. For color images this is 3
nz = 100 # Size of z latent vector (i.e. size of generator input)
ngf = 64 # Size of feature maps in generator
ndf = 64 # Size of feature maps in discriminator
lr = 0.0002 # Learning rate for torch.optimizers
beta1 = 0.5 # Beta1 hyperparam for Adam torch.optimizers
# Number of GPUs available. Use 0 for CPU mode.
ngpu = 8

def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        nn.init.normal_(m.weight.data, 0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        nn.init.normal_(m.weight.data, 1.0, 0.02)
        nn.init.constant_(m.bias.data, 0)

class Generator(nn.Module):
    def __init__(self, ngpu):
        super(Generator, self).__init__()
        self.ngpu = ngpu
        self.main = nn.Sequential(
            # input is Z, going into a convolution
            nn.ConvTranspose2d( nz, ngf * 8, 4, 1, 0, bias=False),
            nn.BatchNorm2d(ngf * 8),
            nn.ReLU(True),
            # state size. (ngf*8) x 4 x 4
            nn.ConvTranspose2d(ngf * 8, ngf * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf * 4),
            nn.ReLU(True),
            # state size. (ngf*4) x 8 x 8
            nn.ConvTranspose2d( ngf * 4, ngf * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf * 2),
            nn.ReLU(True),
            # state size. (ngf*2) x 16 x 16
            nn.ConvTranspose2d( ngf * 2, ngf, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ngf),
            nn.ReLU(True),
            # state size. (ngf) x 32 x 32
            nn.ConvTranspose2d( ngf, nc, 4, 2, 1, bias=False),
            nn.Tanh() # GANHACK #1
            # state size. (nc) x 64 x 64
        )

    def forward(self, input):
        return self.main(input)

class Discriminator(nn.Module):
    def __init__(self, ngpu):
        super(Discriminator, self).__init__()
        self.ngpu = ngpu
        self.main = nn.Sequential(
            # input is (nc) x 64 x 64
            nn.Conv2d(nc, ndf, 4, 2, 1, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # state size. (ndf) x 32 x 32
            nn.Conv2d(ndf, ndf * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ndf * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # state size. (ndf*2) x 16 x 16
            nn.Conv2d(ndf * 2, ndf * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ndf * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # state size. (ndf*4) x 8 x 8
            nn.Conv2d(ndf * 4, ndf * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(ndf * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # state size. (ndf*8) x 4 x 4
            nn.Conv2d(ndf * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid()
        )

    def forward(self, input):
        return self.main(input)


class GAN:
    def __init__(self):
        self.G = Generator(ngpu).to(device)
        # Handle multi-gpu if desired
        if (device.type == 'cuda') and (ngpu > 1):
            self.G = nn.DataParallel(self.G, list(range(ngpu)))

        # Apply the weights_init function to randomly initialize all weights
        #  to mean=0, stdev=0.2.
        self.G.apply(weights_init)

        # Print the model
        print(self.G)

        # Create the Discriminator
        self.D = Discriminator(ngpu).to(device)

        # Handle multi-gpu if desired
        if (device.type == 'cuda') and (ngpu > 1):
            self.D = nn.DataParallel(self.D, list(range(ngpu)))

        # Apply the weights_init function to randomly initialize all weights
        #  to mean=0, stdev=0.2.
        self.D.apply(weights_init)

        # Print the model
        print(self.D)


    def sample(self):
        noise = torch.randn(b_size, nz, 1, 1, device=device)
        # Generate fake image batch with G
        fake = self.G(noise)
        return fake

    def fit(self, epochs=100):
        # Initialize BCELoss function
        criterion = nn.BCELoss()

        # Create batch of latent vectors that we will use to visualize
        #  the progression of the generator
        fixed_noise = torch.randn(64, nz, 1, 1, device=device)

        # Establish convention for real and fake labels during training
        real_label = 1.
        fake_label = 0.
        label_noise = 0.3

        # Setup Adam torch.optimizers for both G and D
        torch.optimizerD = torch.optim.Adam(self.D.parameters(), lr=lr, betas=(beta1, 0.999))
        torch.optimizerG = torch.optim.Adam(self.G.parameters(), lr=lr, betas=(beta1, 0.999))
        
        img_list = []
        G_losses = []
        D_losses = []
        iters = 0

        print("Starting Training Loop...")
        for epoch in range(epochs): # For each epoch
            for i, data in enumerate(dataloader, 0): # For each batch in the dataloader

                ############################
                # (1) Update D network: maximize log(D(x)) + log(1 - D(G(z)))
                ###########################
                ## Train with all-real batch
                self.D.zero_grad()
                real_cpu = data[0].to(device) # Format batch
                b_size = real_cpu.size(0)
                label = torch.full((b_size,), real_label, dtype=torch.float, device=device)
                label += (torch.rand_like(label) * label_noise) - (label_noise / 2) # GANHACK #6 soft & noisy labels
                output = self.D(real_cpu).view(-1) # Forward pass real batch through D # GANHACK #4
                errD_real = criterion(output, label) # Calculate loss on all-real batch
                errD_real.backward() # Calculate gradients for D in backward pass
                D_x = output.mean().item()

                ## Train with all-fake batch
                noise = torch.randn(b_size, nz, 1, 1, device=device) # Generate batch of latent vectors
                fake = self.G(noise) # Generate fake image batch with G
                label.fill_(fake_label)
                label += (torch.rand_like(label) * label_noise) - (label_noise / 2) # GANHACK #6 soft & noisy labels
                
                output = self.D(fake.detach()).view(-1) # Classify all fake batch with D # GANHACK #4 = different discriminator loss
                errD_fake = criterion(output, label) # Calculate D's loss on the all-fake batch # GANHACK #2
                errD_fake.backward() # Calculate the gradients for this batch
                D_G_z1 = output.mean().item()
                
                errD = errD_real + errD_fake # Add the gradients from the all-real and all-fake batches
                torch.optimizerD.step() # Update D

                ############################
                # (2) Update G network: maximize log(D(G(z)))
                ###########################
                self.G.zero_grad()
                label.fill_(real_label)  # fake labels are real for generator cost
                # Since we just updated D, perform another forward pass of all-fake batch through D
                output = self.D(fake).view(-1)
                # Calculate G's loss based on this output
                errG = criterion(output, label)
                # Calculate gradients for G
                errG.backward()
                D_G_z2 = output.mean().item()
                # Update G
                torch.optimizerG.step()

                # Output training stats
                if i % 5 == 0:
                    print('[%d/%d][%d/%d]\tLoss_D: %.4f\tLoss_G: %.4f\tD(x): %.4f\tD(G(z)): %.4f / %.4f'
                        % (epoch, epochs, i, len(dataloader),
                            errD.item(), errG.item(), D_x, D_G_z1, D_G_z2))

                # Save Losses for plotting later
                G_losses.append(errG.item())
                D_losses.append(errD.item())

                # Check how the generator is doing by saving G's output on fixed_noise
                if (iters % 500 == 0) or ((epoch == epochs-1) and (i == len(dataloader)-1)):
                    with torch.no_grad():
                        fake = self.G(fixed_noise).detach().cpu()
                    imgs = torchvision.utils.make_grid(fake, padding=2, normalize=True)
                    img_list.append(imgs)
                    # SAVE IMGS
                    print('saving fake imgs')
                    if not os.path.exists('fake_imgs'):
                        os.mkdir('fake_imgs')
                    imgs = torchvision.transforms.ToPILImage()(imgs)
                    imgs.save(f'fake_imgs/step-{iters}.jpeg', 'JPEG')

                iters += 1


if __name__ == '__main__':
    dataset = ImageDataset('../../Furniture-StyleGAN/images', download=False)
    print(len(dataset))
    # Number of workers for dataloader
    workers = 2
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=workers)

    # Decide which device we want to run on
    device = torch.device("cuda:0" if (torch.cuda.is_available() and ngpu > 0) else "cpu")

    # Plot some training images
    real_batch = next(iter(dataloader))
    plt.figure(figsize=(8,8))
    plt.axis("off")
    plt.title("Training Images")
    plt.imshow(np.transpose(torchvision.utils.make_grid(real_batch[0].to(device)[:64], padding=2, normalize=True).cpu(),(1,2,0)))

    gan = GAN()
    gan.fit()

    # fig = plt.figure(figsize=(8,8))
    # plt.axis("off")
    # ims = [[plt.imshow(np.transpose(i,(1,2,0)), animated=True)] for i in img_list]
    # ani = animation.ArtistAnimation(fig, ims, interval=1000, repeat_delay=1000, blit=True)

    # HTML(ani.to_jshtml())

# %%

plt.figure(figsize=(10,5))
plt.title("Generator and Discriminator Loss During Training")
plt.plot(G_losses,label="G")
plt.plot(D_losses,label="D")
plt.xlabel("iterations")
plt.ylabel("Loss")
plt.legend()
plt.show()
# %%



import torch
import utils
from NN import NN 
import numpy as np
import torch.nn.functional as F
from ray import tune
import os
from train import train
from torch.utils.data import DataLoader

batch_size = 16

train_data, val_data, test_data = utils.get_splits()

class AutoEncoder(torch.nn.Module):
    def __init__(
            self, 
            encoder_layers, 
            decoder_layers
        ):
        super().__init__()
        self.encoder = NN(encoder_layers)
        self.decoder = NN(decoder_layers)

    def encode(self, x):
        return self.encoder(x)

    def decode(self, z):
        return self.decoder(z)

    def forward(self, x):
        z = self.encode(x)
        h = self.decode(z)
        return h

layers = [784, 256, 64]#, 64, 8, 2]

class AEDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, transform=None):
        self.dataset = dataset
        self.transform = transform

    def __getitem__(self, idx):
        x, _ = self.dataset[idx]
        if self.transform:
            x = self.transform(x)
        return (x, x)

    def __len__(self):
        return len(self.dataset)

def flatten(x):
    return torch.flatten(x)

train_loader = DataLoader(AEDataset(train_data, transform=flatten), shuffle=True, batch_size=batch_size)
val_loader = DataLoader(AEDataset(val_data, transform=flatten), shuffle=True, batch_size=batch_size)
test_loader = DataLoader(AEDataset(test_data, transform=flatten), shuffle=True, batch_size=batch_size)

class FlatAE(torch.nn.Module):
    def __init__(self,
        encoder_layers,
        decoder_layers
    ):
        super().__init__()
        self.layers = torch.nn.Sequential(
            torch.nn.Flatten(),
            AutoEncoder(
                encoder_layers,
                decoder_layers
            )
        )

    def forward(self, x):
        return self.layers(x)

# ae = FlatAE()

def sample(writer, device='cpu'):
    z = torch.randn(batch_size, latent_vec_size).to(device)
    for img in G(z):
        writer.add_image(f'test', img)

def visualise_reconstruction(writer, originals, reconstructions):

    writer.add_images('originals', originals)
    writer.add_images('reconstructions', reconstructions)

for idx in range(1, 9):
    layers = [
        784,
        *[2**(9-idx) for idx in range(idx)]
    ]
    print(layers)
    model = FlatAE(
        encoder_layers=layers,
        decoder_layers=layers[::-1]
    )
    print(f'training model {idx}')
    model, writer = train(
        model=model, 
        logdir='Autoencoder',
        train_loader=train_loader,
        val_loader=val_loader,
        test_loader=test_loader,
        loss_fn=F.mse_loss,
        epochs=10
    )
    for batch in train_loader:
        x, _ = batch
        reconstructions = model(x)
        x = x.view(x.shape[0], 1, 28, 28)
        reconstructions = reconstructions.view(reconstructions.shape[0], 1, 28, 28)
        visualise_reconstruction(writer, x, reconstructions)
        break



import torch

class VerboseLayer(torch.nn.Module):
    def __init__(self, layer_idx):
        super().__init__()
        self.layer_idx = layer_idx

    def forward(self, x):
        print(f'layer: {self.layer_idx}\tshape: {x.shape}')
        return x


import torch
class NN(torch.nn.Module):
    def __init__(self, layer_widths):
        super().__init__()
        l = []

        n_layers = len(layer_widths)

        for idx in range(n_layers - 1):
            l.append(torch.nn.Linear(layer_widths[idx], layer_widths[idx+1]))   # add a linear layer
            if idx + 1 != n_layers: # if this is not the last layer ( +1 = zero indexed) (-1 = layer b4 last)
                l.append(torch.nn.ReLU())   # activate
        self.layers = torch.nn.Sequential(
            *l
        )

    def forward(self, x):
        x = self.layers(x)
        return x